 /*
  *
  * Piecewise-constant file format.
  *
  * Paul Ausbeck 4/18/98
  */

#define _INC_IFILELIB

typedef int (*PUT_FUNC)(int,void *);
typedef int (*GET_FUNC)(void *);
typedef long (*TELL_FUNC)(void *);

class ifile_io_wrapper
{
  void *ioptr;
  union
  {
    int (*putbyte)(int value, void *ioptr);
    int (*getbyte)(void *ioptr);
  };

public:
  
  ifile_io_wrapper(void *ioptr, int (*const putbyte)(int value, void *ioptr))
  {
    this->ioptr = ioptr;
    this->putbyte = putbyte;
  }
  ifile_io_wrapper(void *ioptr, int (*const getbyte)(void *ioptr))
  {
    this->ioptr = ioptr;
    this->getbyte = getbyte;
  }

  int get(void) { return (*getbyte)(ioptr); }
  int put(int value) { return (*putbyte)(value, ioptr); }

  int read(void *buffer, int bytes)
  {
     for (char *ptr = (char *) buffer; ptr < (char *) buffer + bytes; )
     {
       int byte = (*getbyte)(ioptr);
       if (byte != -1)
         *ptr++ = byte;
       else
         break;
     }

     return ptr - (char *) buffer;
  }
  int write(void *buffer, int bytes)
  {
     for (char *ptr = (char *) buffer; ptr < (char *) buffer + bytes; 
          (*putbyte)(*ptr++, ioptr));

     return bytes;
  }

  void *iobuf(void) { return ioptr; }
  int (*const putfunc(void))(int value, void *ioptr) { return putbyte; }
  int (*const getfunc(void))(void *ioptr) { return getbyte; }
};

enum ifile_header_constants
{
  IFILE_MAGIC = 'a' << 24 | 'V' << 16 | 'y' << 8 | '4',

  IFILE_USES_PALETTE = 0x1,
  IFILE_USES_COLOR = 0x2,
  IFILE_USES_ALPHA = 0x4,
  IFILE_USES_PWC = 0x3,
  IFILE_USES_GRAY = 0x0
};

class ifile_header
{
 struct _header
  {
    int magic;

    unsigned short rows;
    unsigned short columns;

    unsigned char depth;

    unsigned char uses : 3;
    unsigned char inverse_row_order : 1;
    unsigned char reserved1 : 4;

    unsigned short reserved2;

    int check(void)
    {
      return magic != IFILE_MAGIC;
    }
  } header;

  public:

  void init(int rows = 0, int columns = 0, int depth = 0, int uses = 0,
            int inverse_row_order = 0)
  {
    header.magic = IFILE_MAGIC;
    header.rows = rows;
    header.columns = columns;
    header.depth = depth;
    header.uses = uses;
    header.inverse_row_order = inverse_row_order;
    header.reserved1 = 0;
    header.reserved2 = 0;
    header.magic = IFILE_MAGIC;
  }

  ifile_header(void) { init(); }
  ifile_header(int rows, int columns, int depth, int uses, int inverse_row_order = 0)
  {
    init(rows, columns, depth, uses, inverse_row_order);
  }
  ifile_header(ifile_io_wrapper *file) { read(file); }

  int check(void) { return header.check(); }

  int read(ifile_io_wrapper *file)
  {
     int length = file->read(this, sizeof(*this));

     return length == sizeof(*this) && ! header.check();
  }

  int write(ifile_io_wrapper *file)
  {
     int length = file->write(this, sizeof(*this));

     return length == sizeof(*this) && !header.check();
  }

  int rows(void) { return header.rows; }
  int columns(void) { return header.columns; }
  int depth(void) { return header.depth; }

  int uses(void) { return header.uses; }
  int uses_palette(void) { return header.uses & IFILE_USES_PALETTE; }
  int uses_color(void) { return header.uses & IFILE_USES_COLOR; }
  int uses_alpha(void) { return header.uses & IFILE_USES_ALPHA; }

  int inverted_rows(void) { return header.inverse_row_order; }
};

enum ifile_chunks
{
  IFILE_CHUNK_PALETTE = 0x0,
  IFILE_CHUNK_PWS = 0x1,
  IFILE_CHUNK_PWC = 0x2
};

class ifile_chunk_header
{
  unsigned char _type;

public:
  ifile_chunk_header(int type) { _type = type; }
  ifile_chunk_header(ifile_io_wrapper *file)
  {
    _type = file->get();
  }
  int read(ifile_io_wrapper *file) 
  {
    _type = file->get();
    return 0;
  }

  int write(ifile_io_wrapper *file) 
  {
    file->put(_type);
    return 0;
  }
};

struct ifile_palette_entry
{
  unsigned char blue; 
  unsigned char green; 
  unsigned char red; 
  unsigned char unused; 
  ifile_palette_entry &operator =(ifile_palette_entry &what)
  {
    red = what.red; green = what.green; blue = what.blue;
    return *this;
  }
  void set(int to) { red = green = blue = to; }
};

class palette_chunk
{
  int entries;
  ifile_palette_entry *palette;

  int index;
  int bgr;

public:
  void init(void) { entries = 0; palette = 0; index = 0; bgr = 0; }
  palette_chunk(void) { init(); }
 
  palette_chunk(int entries, ifile_palette_entry *pal) 
  { 
    this->entries = entries;
    palette = new ifile_palette_entry[entries];

    for (int i = 0; i < entries; i++)
      palette[i] = pal[i];

    index = 0; bgr = 0;
  }

  ~palette_chunk(void)
  {
    if (palette)
      delete[] palette;

    palette = 0;
    entries = 0;
  }

  int colors(void) { return entries; }
  ifile_palette_entry *color_array(void) { return palette; }

  int complete(void) { return index == entries; }
  void read(ifile_io_wrapper *file) // re-entrant
  {
    if (palette == 0)
    { /* Assume data is available for read, method fails if not. */
      entries = 0;
      file->read(&entries, 2);
      palette = new ifile_palette_entry[entries];
      for (int i = 0; i < entries; i++)
        palette[i].set(0);
    }

    if (entries <= 256) // safety
    {
      for (; index < entries; bgr = 0, index++)
      {
        for(; bgr < 3; bgr++)
        {
          if (!file->read(((unsigned char *) (palette + index)) + bgr, 1))
            return;
        }
      }
    }
  }

  void read(ifile_io_wrapper *file, int depth, int uses)
  {
    if (uses & IFILE_USES_PALETTE)
      read(file);
    else
      linear(depth);
  }

  void linear(int depth)
  {
    entries = 1 << depth;
    palette = new ifile_palette_entry[entries];

    if (depth == 8)
    {
      for (int i = 0; i < entries; i++)
        palette[i].set(i);
    }
    else if(depth == 4)
    {
      for (int i = 0, j = 8; i < entries; i++, j += 16)
        palette[i].set(j);
    }
    else
    {
      palette[0].set(255);
      palette[1].set(0);
    }
  }

  void write(ifile_io_wrapper *file, int palette_entries)
  {
    palette_entries = palette_entries ? palette_entries : entries;

    file->write(&palette_entries, 2);

    for (int index = 0; index < palette_entries; index++)
    {
      file->write(&palette[index], 3);
    }
  }
};
